@@ -38,6 +38,7 @@ typedef struct {
 	int wcmMTChannel;
 	int wcmEventCnt;
 	struct input_event wcmEvents[MAX_USB_EVENTS];
+	uint32_t wcmEventFlags;      /* event types received in this frame */
 	int nbuttons;                /* total number of buttons */
 	int npadkeys;                /* number of pad keys in the above array */
 	int padkey_code[WCM_MAX_BUTTONS];/* hardware codes for buttons */
@@ -59,6 +60,8 @@ static void usbParseEvent(InputInfoPtr pInfo,
 	const struct input_event* event);
 static void usbParseSynEvent(InputInfoPtr pInfo,
 			     const struct input_event *event);
+static void usbParseMscEvent(InputInfoPtr pInfo,
+			     const struct input_event *event);
 static void usbDispatchEvents(InputInfoPtr pInfo);
 static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int serial);
 
@@ -234,6 +237,9 @@ static struct WacomModelDesc
 	{ WACOM_VENDOR_ID, 0xF0, 100000, 100000, &usbCintiq,     "DTU1631"		},
 	{ WACOM_VENDOR_ID, 0x35a, 100000, 100000, &usbCintiq,    "DTH1152"              },
 	{ WACOM_VENDOR_ID, 0x368, 100000, 100000, &usbCintiq,    "DTH1152"              }, /* Touch */
+	{ WACOM_VENDOR_ID, 0x382, 100000, 100000, &usbCintiq,    "DTK2451"              },
+	{ WACOM_VENDOR_ID, 0x37D, 100000, 100000, &usbCintiq,    "DTH2452"              },
+	{ WACOM_VENDOR_ID, 0x37E, 100000, 100000, &usbCintiq,    "DTH2452"              }, /* Touch */
 
 	{ WACOM_VENDOR_ID, 0x41, 100000, 100000, &usbIntuos2,    "Intuos2 4x5"		},
 	{ WACOM_VENDOR_ID, 0x42, 100000, 100000, &usbIntuos2,    "Intuos2 6x8"		},
@@ -790,8 +796,8 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 			private->wcmPenTouch = TRUE;
 	}
 
-	/* A generic protocol device does not report ABS_MISC event */
-	if (!ISBITSET(abs, ABS_MISC))
+	/* Non-wacom devices, and Wacom devices without an ABS_MISC should be treated as generic */
+	if (common->vendor_id != WACOM_VENDOR_ID || !ISBITSET(abs, ABS_MISC))
 		common->wcmProtocolLevel = WCM_PROTOCOL_GENERIC;
 
 	if (ioctl(pInfo->fd, EVIOCGBIT(EV_SW, sizeof(sw)), sw) < 0)
@@ -929,7 +935,8 @@ static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int
 			if (i == PAD_CHANNEL)
 				continue;
 
-			if (!common->wcmChannel[i].work.proximity)
+			if (!common->wcmChannel[i].work.proximity &&
+			    !common->wcmChannel[i].valid.state.proximity)
 			{
 				channel = i;
 				memset(&common->wcmChannel[channel],0, sizeof(WacomChannel));
@@ -967,6 +974,13 @@ static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int
 	return channel;
 }
 
+static inline void
+usbResetEventCounter(wcmUSBData *private)
+{
+	private->wcmEventCnt = 0;
+	private->wcmEventFlags = 0;
+}
+
 static void usbParseEvent(InputInfoPtr pInfo,
 	const struct input_event* event)
 {
@@ -976,24 +990,59 @@ static void usbParseEvent(InputInfoPtr pInfo,
 
 	DBG(10, common, "\n");
 
-	/* store events until we receive the MSC_SERIAL containing
-	 * the serial number or a SYN_REPORT.
-	 */
+	/* store events until we receive a SYN_REPORT */
 
 	/* space left? bail if not. */
 	if (private->wcmEventCnt >= ARRAY_SIZE(private->wcmEvents))
 	{
 		LogMessageVerbSigSafe(X_ERROR, 0, "%s: usbParse: Exceeded event queue (%d) \n",
 				      pInfo->name, private->wcmEventCnt);
-		private->wcmEventCnt = 0;
+		usbResetEventCounter(private);
 		return;
 	}
 
 	/* save it for later */
 	private->wcmEvents[private->wcmEventCnt++] = *event;
+	private->wcmEventFlags |= 1 << event->type;
 
-	if (event->type == EV_MSC || event->type == EV_SYN)
-		usbParseSynEvent(pInfo, event);
+	switch (event->type)
+	{
+		case EV_MSC:
+			usbParseMscEvent(pInfo, event);
+			break;
+		case EV_SYN:
+			usbParseSynEvent(pInfo, event);
+			break;
+		default:
+			break;
+	}
+}
+
+static void usbParseMscEvent(InputInfoPtr pInfo,
+			     const struct input_event *event)
+{
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
+	WacomCommonPtr common = priv->common;
+	wcmUSBData* private = common->private;
+
+	if (event->code != MSC_SERIAL)
+		return;
+
+	if (event->value != 0)
+	{
+		/* save the serial number so we can look up the channel number later */
+		private->wcmLastToolSerial = event->value;
+	}
+	else
+	{
+		/* we don't report serial numbers for some tools but we never report
+		 * a serial number with a value of 0 - if that happens drop the
+		 * whole frame */
+		LogMessageVerbSigSafe(X_ERROR, 0,
+				      "%s: usbParse: Ignoring event from invalid serial 0\n",
+				      pInfo->name);
+		usbResetEventCounter(private);
+	}
 }
 
 /**
@@ -1008,46 +1057,24 @@ static void usbParseSynEvent(InputInfoPtr pInfo,
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 	wcmUSBData* private = common->private;
+	const uint32_t significant_event_types = ~(1 << EV_SYN | 1 << EV_MSC);
 
-	if ((event->type == EV_MSC) && (event->code == MSC_SERIAL))
-	{
-		/* we don't report serial numbers for some tools
-		 * but we never report a serial number with a value of 0 */
-		if (event->value == 0)
-		{
-			LogMessageVerbSigSafe(X_ERROR, 0,
-					      "%s: usbParse: Ignoring event from invalid serial 0\n",
-					      pInfo->name);
-			goto skipEvent;
-		}
-
-		/* save the serial number so we can look up the channel number later */
-		private->wcmLastToolSerial = event->value;
-
+	if (event->code != SYN_REPORT)
 		return;
 
-	} else if ((event->type == EV_SYN) && (event->code == SYN_REPORT))
-	{
-		/* end of record. fall through to dispatch */
-	}
-	else
-	{
-		/* not an SYN_REPORT and not an SYN_REPORT, bail out */
-		return;
-	}
-
 	/* ignore events without information */
 	if ((private->wcmEventCnt < 2) && private->wcmLastToolSerial)
 	{
-		DBG(3, common, "%s: dropping empty event"
-			" for serial %d\n", pInfo->name,
-			private->wcmLastToolSerial);
+		DBG(3, common, "%s: dropping empty event for serial %d\n",
+		    pInfo->name, private->wcmLastToolSerial);
 		goto skipEvent;
 	}
 
-	/* ignore sync windows that contain no data */
-	if (private->wcmEventCnt == 1 &&
-	    private->wcmEvents->type == EV_SYN) {
+
+	/* If all we get in an event frame is EV_SYN/EV_MSC, we don't have
+	 * real data to process. */
+	if ((private->wcmEventFlags & significant_event_types) == 0)
+	{
 		DBG(6, common, "no real events received\n");
 		goto skipEvent;
 	}
@@ -1056,7 +1083,7 @@ static void usbParseSynEvent(InputInfoPtr pInfo,
 	usbDispatchEvents(pInfo);
 
 skipEvent:
-	private->wcmEventCnt = 0;
+	usbResetEventCounter(private);
 }
 
 static int usbFilterEvent(WacomCommonPtr common, struct input_event *event)
@@ -1183,7 +1210,15 @@ static int usbFindDeviceTypeById(int device_id)
 	}
 }
 
-static void usbParseAbsEvent(WacomCommonPtr common,
+/**
+ * Handle ABS events strictly according to their definition as documented
+ * in the Linux kernel.
+ *
+ * @param common
+ * @param event
+ * @param channel_number
+ */
+static int usbParseGenericAbsEvent(WacomCommonPtr common,
 			    struct input_event *event, int channel_number)
 {
 	WacomChannel *channel = &common->wcmChannel[channel_number];
@@ -1198,12 +1233,6 @@ static void usbParseAbsEvent(WacomCommonPtr common,
 		case ABS_Y:
 			ds->y = event->value;
 			break;
-		case ABS_RX:
-			ds->stripx = event->value;
-			break;
-		case ABS_RY:
-			ds->stripy = event->value;
-			break;
 		case ABS_RZ:
 			ds->rotation = event->value;
 			break;
@@ -1224,6 +1253,39 @@ static void usbParseAbsEvent(WacomCommonPtr common,
 		case ABS_WHEEL:
 			ds->abswheel = event->value;
 			break;
+		case ABS_THROTTLE:
+			ds->throttle = event->value;
+			break;
+		default:
+			change = 0;
+	}
+
+	return change;
+}
+
+/**
+ * Handle ABS events according to Wacom (protocol 4/5) conventions. Only
+ * those events which are not interpreted in the generic manner need to be
+ * handled here.
+ *
+ * @param common
+ * @param event
+ * @param channel_number
+ */
+static int usbParseWacomAbsEvent(WacomCommonPtr common,
+			    struct input_event *event, int channel_number)
+{
+	WacomChannel *channel = &common->wcmChannel[channel_number];
+	WacomDeviceState *ds = &channel->work;
+	int change = 1;
+
+	switch(event->code) {
+		case ABS_RX:
+			ds->stripx = event->value;
+			break;
+		case ABS_RY:
+			ds->stripy = event->value;
+			break;
 		case ABS_Z:
 			ds->abswheel = event->value;
 			break;
@@ -1232,8 +1294,6 @@ static void usbParseAbsEvent(WacomCommonPtr common,
 			if ((common->vendor_id == WACOM_VENDOR_ID) &&
 			    (common->tablet_id == 0xF4 || common->tablet_id == 0xF8))
 				ds->abswheel2 = event->value;
-			else
-				ds->throttle = event->value;
 			break;
 		case ABS_MISC:
 			ds->proximity = (event->value != 0);
@@ -1244,6 +1304,28 @@ static void usbParseAbsEvent(WacomCommonPtr common,
 			change = 0;
 	}
 
+	return change;
+}
+
+/**
+ * Handle an incoming ABS event.
+ *
+ * @param common
+ * @param event
+ * @param channel_number
+ */
+static void usbParseAbsEvent(WacomCommonPtr common,
+			    struct input_event *event, int channel_number)
+{
+	WacomChannel *channel = &common->wcmChannel[channel_number];
+	WacomDeviceState *ds = &channel->work;
+	Bool change;
+
+	change = usbParseGenericAbsEvent(common, event, channel_number);
+	if (common->wcmProtocolLevel != WCM_PROTOCOL_GENERIC) {
+		change |= usbParseWacomAbsEvent(common, event, channel_number);
+	}
+
 	ds->time = (int)GetTimeInMillis();
 	channel->dirty |= change;
 }
@@ -1581,7 +1663,8 @@ static int deviceTypeFromEvent(WacomCommonPtr common, int type, int code, int va
 			case ABS_MT_TRACKING_ID:
 				return TOUCH_ID;
 			case ABS_MISC:
-				return usbFindDeviceTypeById(value);
+				if (common->wcmProtocolLevel != WCM_PROTOCOL_GENERIC)
+					return usbFindDeviceTypeById(value);
 		}
 	}
 
